"""

purpose: assorted function, which are a bit untested


Functions:
----------
importDataFromFolder - IF the user already has a data folder with certain
specifications
>> from supplychainmodelhelper import datahandling as dh
>> help(dh.importDataFromFolder)
this function may be used to create a hdf5 database without much effort

decBSList - if a list of strings or bytestrings are given, this function
returns a list of strings. No need to check if some string is one or the
other, just apply this to your list.

decBS - if a string or bytestring is given, this function returns a string.
No need to check if some string is one or the other, just apply this to your
list.
"""

import csv #1.0
import h5py # 2.10.0
import pandas as pd # 0.25.2
import os
import warnings
warnings.filterwarnings("ignore")



def decBSList(inputList: list) -> list:
    """
    purpose:convert list of bytestrings to list of strings
    if it is already a string, do nothing

    :param inputList: a list of strings to be converted
    (may be str or bytestring)

    :return: list of strings converted to str

    example:
    from supplychainmodelhelper import datahandling as dh
    blist = [b'test',b'another test',b'not another test']
    strlist = dh.decBSlist(blist)

    """
    definiteList = []
    for i in range(len(inputList)):  
        try:
            definiteList.append(inputList[i].decode())
        except (UnicodeDecodeError, AttributeError):
            definiteList.append(inputList[i])
   
    return definiteList


def decBS(inputString: str) -> str:
    """
    purpose: convert list of bytestrings to list of strings
    if it is already a string, do nothing

    :param inputString: a str or bytestring

    :return: string

    example:
    from supplychainmodelhelper import datahandling as dh
    blist = b'test'
    strlist = dh.decBSlist(blist)

    """
    try:
        definiteString = inputString.decode()
    except (UnicodeDecodeError, AttributeError):
        definiteString = inputString
   
    return definiteString


def importDataFromFolder(
        fileOutputName: str,
        fileInputPath: str,
        fileInMetadataName='folderMD.csv'
) -> bool:
    """
    purpose: import files from existing folder structure
    folder needs to consists of folderMD.csv:
        - minimum entries for headers of folderMD.csv:
            'Folder', 'ID', 'Title', 'Description' optional to enter more
            categories, will be automatically integrated into database
        - fill this csv file with naming the data containing folders,
            give unique IDs and proper title and description
        - every subfolder mentioned in folderMD.csv should contain another
            csv file named as the subfolder (e.g. subfolder is 'Test',
            so the csv file in 'Test' should be named 'TestMD.csv')
        - minimum entries for headers of 'subfolder'MD.csv:
            'Dataset',
            'Filename',
            'Format',
            'ID',
            'Title',
            'Rows',
            'Columns',
            'Encoding'
            ('Dataset' is the identifier you give when retrieving dataset
            from db)
        - put file with 'Filename' into this subfolder, this file will be
            added to database (with header!)

    TODO maybe an example folder?
    current implementation: 1D / 2D dataset only

    :param fileOutputPath:
    :param fileInputPath: the path of the folder 
    :param fileInMetadataName(optional): filename of folderMD.csv(default), maybe different filename
    
    :return: true if input file existed and consistent with all data in folder

    example: 
    from supplychainmodelhelper import datahandling as dh
    path2DataFolder = '../data/'#folder where your data set lies
    myfname = 'test.hdf5'

    didItWork4 = dh.importDataFromFolder(
        fileOutputName=myfname,
        fileInputPath=path2DataFolder
    )
    if didItWork4:
        print('yes!')
    else:
        print('no!')

    """
    done = []
    

    done.append(True) \
        if isinstance(fileInputPath, str) \
        else done.append(False)
    done.append(True) \
        if isinstance(fileInMetadataName, str) \
        else done.append(False)
    done.append(True) \
        if isinstance(fileOutputName, str) \
        else done.append(False)

    # start reading metadata structure
    folderMD = csv.DictReader(
        open(fileInputPath + fileInMetadataName),
        delimiter=';'
    )
    done.append(True) \
        if(os.path.exists(fileInputPath + fileInMetadataName)) \
        else done.append(False)
    
    listOfContent = []
    # writing the hdf5 structure
    with h5py.File(fileOutputName,'w') as db:
        ####################################################
        # creating the structure from meta data schema table
        ####################################################
        # getting the metadata from each metadata table in
        # each of the folders as described in fileFolderLoc
        for row in folderMD:
            listOfContent.append(row['Folder'])
            myAttrs = list(row.keys())        
            # creating the folders and look out for subfolders
            if not row['Subfolder']:
                db.create_group(row['Folder'])
                for run in myAttrs:
                    # attach all info to metadata schema
                    if run != 'Folder' and run != 'Subfolder':
                        db[row['Folder']].attrs[run] = row[run]

            # creating the subfolders(aka subgroups) inside hdf5 db
            # if existing in metadata file
            if row['Subfolder']:
                db[row['Folder']].create_group(row['Subfolder'])

                for run in myAttrs:
                    if run != 'Folder' and run != 'Subfolder':
                        db[
                            row['Folder'] +
                            '/' +
                            row['Subfolder']
                        ].attrs[run] = row[run]
        
    ####################################################
    # incorporating the datasets with corresponding metadata
    ####################################################
        for foldername in db:
    
            #print(foldername)
            # code assumes
            #   that meta datafile is located in file folder and
            #   last 2 letters of filename is "MD" and
            #   is in csv-format
            metaDataFile = csv.DictReader(
                open(
                    fileInputPath +
                    foldername +
                    "/" +
                    foldername +
                    "MD.csv"
                ),
                delimiter=';'
            )
            done.append(True) if(os.path.exists(
                fileInputPath +
                foldername +
                "/" +
                foldername +
                "MD.csv")
            ) else done.append(False)
            
            for row in metaDataFile:
            #     # read in the data set:
            #     NEEDS to have index and column header in csv file
            #     #print(row)
                dataSet = pd.read_csv(
                    fileInputPath + foldername + "/" + row['Filename'],
                    header=0,
                    sep=';',
                    index_col=0,
                    encoding=row['Encoding'],
                    dtype='str'
                )
                done.append(True) if(os.path.exists(
                    fileInputPath+foldername+"/"+row['Filename'])
                ) else done.append(False)
                dataSet = dataSet.fillna(value=0)

            #     # create the hdf5 dataset and store it in the hdf5-file
                dt = h5py.string_dtype(encoding='utf-8')
                d5 = db[foldername].create_dataset(
                    row['Dataset'],
                    data=dataSet,
                    dtype=dt
                )
                
                # attach metadata information to newly created data set
                myAttrs = list(row.keys())
                for run in myAttrs:
                    if run != 'Dataset':
                        db[
                            foldername+
                            '/'+
                            row['Dataset']
                        ].attrs[run] = row[run]
                

                # attach dimension scale from csv file,
                # ASSUMING it's always a 2d table...
                # TODO make it dynamical, so that the reader recognizes
                #  the format!
                # has a unique ID, given that in the MD-files
                # this ID is unique (filewise), as well!
                # TODO currently unique ID is given in filename
                #  --> make it internal UID!
                try:
                    dtIndex = 'i'
                    db[foldername].create_dataset(
                        "ScaleRowOfDataSet"+row['ID'],
                        data=dataSet.index.astype(int).values,
                        dtype=dtIndex
                    )
                except TypeError:
                    # this only works for py3!
                    dtIndex = h5py.string_dtype(encoding='utf-8')
                    db[foldername].create_dataset(
                        "ScaleRowOfDataSet"+row['ID'],
                        data=dataSet.index.values,
                        dtype=dtIndex
                    )
        
                # attach the header row and header columne of the
                # table as meta data info
                # TODO automatic recognition of header
                #  -> cell data type analysis (header = string, data=not string)
                d5.dims[0].label = row['Rows']
                d5.dims[0].attach_scale(
                    db[foldername +
                       "/ScaleRowOfDataSet" +
                       row['ID']
                    ]
                )
                

                try:
                    dtCol = 'i'
                    db[foldername].create_dataset(
                        "ScaleColOfDataSet" + row['ID'],
                        data=dataSet.columns.astype(int).values,
                        dtype=dtCol
                    )
                except TypeError:
                    # this only works for py3!
                    dtCol = h5py.string_dtype(encoding='utf-8')
                    db[foldername].create_dataset(
                        "ScaleColOfDataSet" + row['ID'],
                        data=dataSet.columns.values,
                        dtype=dtCol
                    )
                
                # encoding list of strings to list of b-strings (py3 necessity)
                d5.dims[1].label = row['Columns']
                d5.dims[1].attach_scale(
                    db[
                        foldername +
                        "/ScaleColOfDataSet" +
                        row['ID']
                    ]
                )

        # DATABASE IS WRITTEN AND FILE IS CLOSED!
    #done = True
    return all(done)


"""
purpose: simple drawing extension, specialised for supply chain modelling

version 0:
    boundary conditions:
        stages: order of actors are drawn as stages
            example:
            actors = ['producer', 'consumer']
            -> producer are on stage 1, consumer are on stage 2 of this
            2-stages-graph

        assumption: output = input = node size
            the smaller the node the less the output/input of the node

        heatmap for <<chosen edge_label>>(if its a number): with colorbar

        colormap for different products

"""
import itertools
import operator
import warnings

import networkx as nx
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from supplychainmodelhelper import graphoperations as go


class Drawer:
    """
    purpose: draw a supply chain graph
    assumption for standard view (V1):
    layers: actor type defines the number of stages (e.g. producer, warehouse,
    store, consumer)
    color of the edges are heatmap distributed -> log10?(amount)
    - from blue to red
    node size is output/input (possibly log)
    products can only be seen one at a time
    """

    # std_sc_attr = ['actor','location','product']
    # TODO: multiple products

    def __init__(self, SCGraph):
        """
        purpose: initialize drawer with standard view

        :param SCGraph: the graph to be drawn
        """
        # TODO: check if graph has all necessary attributes!

        self.SCGraph = SCGraph
        # for it to work these attributes needed to be defined on the graph
        needed_nodes = ['actor', 'output', 'input']
        needed_edges = 'weight'

        found_nodes = [
            node_type for node_type
            in go.getExistingAttrs(self.SCGraph,gtype='nodes')
            if node_type in needed_nodes
        ]

        # error message if discrepancy is found
        # TODO: this is a hot fix. check for all attributes specifically
        if len(found_nodes) < len(needed_nodes):
            raise Exception(
                "found these node types: " +
                str(found_nodes) +
                ", but need these: " +
                str(needed_nodes)
            )
        if needed_edges not in go.getExistingAttrs(
                self.SCGraph,
                gtype='edges'
        ):
            raise Exception(
                "Please add edge attribute: " +
                needed_edges +
                "to the graph."
            )

    def draw_it(self, product, actor_type_names=None):
        """
        purpose: execute drawing of sc graph
        :param product: choose the name of the product to be shown.
        :param actor_type_names: (optional) list of actor types as defined in
        given graph. Expected to be found in node_attribute "actor".
        choose names of actor types (minimum 2).

        :return: image of graph

        TODO potential parameters:
            product -> multiple products if no product is chosen???
            node bg color = treshold color, size
            node fg color
            treshold value
            other heatmaps for edges
            labeling nodes with locations
            no stages
        """
        # get the ids of the nodes to be drawn
        node_ids_to_draw = go.getListOfNodeIDs(
            SCGraph=self.SCGraph,
            products=product,
            actors=actor_type_names
        )

        # check the given list of actor_type_names
        if actor_type_names:
            if isinstance(actor_type_names, list):
                if len(actor_type_names) < 2:
                    raise Exception(
                        "number of elements in 'actor_type_names' "
                        "should be at least 2!"
                    )
            else:
                raise Exception(
                    "the parameter 'actor_type_names' "
                    "should be a list"
                )

        # prepping for different aspects of the process of drawing
        node_positions, node_stages, node_stage_names = self.node_position(
            filtered_nodes=node_ids_to_draw
        )
        sm, edge_options = self.edge_color()
        node_list, node_size = self.node_size()
        node_size_avg = sum(node_size) / len(node_size)
        node_size_max = max(node_size)
        node_size_min = min(node_size)

        # backgroung nodes drawing
        import math
        fig, ax = plt.subplots()
        nx.draw(self.SCGraph,
                node_positions,
                nodelist=node_list,
                node_color='grey',
                node_size=node_size_avg,
                **edge_options,
                ax=ax
                )

        # foreground nodes drawing with edges
        nx.draw_networkx_nodes(self.SCGraph,
                               node_positions,
                               node_color='k',
                               nodelist=node_list,
                               node_size=node_size
                               )

        # adding actor labels
        ax.tick_params(left=True,
                       bottom=False,
                       labelleft=True,
                       labelbottom=False
                       )

        # set labels for actors
        ax.set_yticks(node_stages)
        ax.set_yticklabels(node_stage_names)

        cbar = plt.colorbar(sm)
        cbar.set_label('transported amount of wares in tons')

        plt.axis('on')

        legend_info = self.legend_node_size(
            math.sqrt(math.pi * node_size_min),
            math.sqrt(math.pi * node_size_avg),
            math.sqrt(math.pi * node_size_max)
        )
        ax.legend(handles=legend_info, loc=9, bbox_to_anchor=(1.3, 0.12))
        ax.get_legend().set_title("throughput")
        plt.show()

    @staticmethod
    def legend_node_size(mymin, myavg, mymax):
        """
        purpose: adjust the size of a node according to output / input of node

        :param mymin: node size minimum in area px
        :param myavg: node size average in area px
        :param mymax: node size maximum in area px

        :return: the legend with all chosen colors, markers, sizes and labels
        """
        mylegend = [
            Line2D([0], [0],
                   markerfacecolor='k',
                   marker='o',
                   markersize=mymin + 1,
                   label='min',
                   color='w'
                   ),
            Line2D([0], [0],
                   markerfacecolor='gray',
                   marker='o',
                   markersize=myavg,
                   label='avg',
                   color='w'
                   ),
            Line2D([0], [0],
                   markerfacecolor='k',
                   marker='o',
                   markersize=mymax,
                   label='max',
                   color='w'
                   )
        ]
        return mylegend

    def node_position(self, node_attr='actor', filtered_nodes=None):
        """
        purpose: return node position on several stages based on content of
        graph(node_attr)

        standard: actors of the same type on the same stage (in the order as
        defined by user)

        :param node_attr: (optional, default: actor) attribute of nodes in
        graph that should be displayed. If ignored, the actors are shown.
        :param filtered_nodes: (optional, default: None) input for list of
        specified node IDs. In ignored, all nodes are chosen.

        :return: position of nodes for drawing: dict(node_ids:(xcoord,ycoord))

        TODO: node_attr - what if other attribute than "actor" is chosen?
            use case (product): show interaction between food processors (e.g.
            sugar beets -> sugar -> pastries)
            use case (location): show how central a node is
        TODO: current layout: grid (senders top, receivers bottom) --> new
        layouts?
        """

        # list of tuples with node ids and actor content
        node_ids_with_actors = go.getNodeIDswAttr(
            SCGraph=self.SCGraph,
            nameOfAttr=node_attr,
            listOfNodeIDs=filtered_nodes
        )

        # list of lists, ordered by content of actor
        # each element of the list is a list with the same type of actor

        stagewise_actors = [
            list(group) for _, group
            in itertools.groupby(
                node_ids_with_actors,
                operator.itemgetter(1)
            )
        ]
        nr_of_stages = len(stagewise_actors)
        nr_of_layers = nr_of_stages - 1

        # check if edges between stages contain attribute 'weight'
        for layer in range(nr_of_layers):
            sender_nodes = [i[0] for i in stagewise_actors[layer]]
            receiver_nodes = [i[0] for i in stagewise_actors[layer + 1]]
            comb_sender, comb_receiver = go.getAllCombinations(
                sender_nodes,
                receiver_nodes
            )
            edges_to_be_drawn = go.convertTup2LoT(comb_sender, comb_receiver)

            existing_edges_in_layer = go.getEdgeIDswAttr(
                SCGraph=self.SCGraph,
                attr='weight'
            )
            sid, rid, _ = go.convertLoT2NTs(existing_edges_in_layer)
            existing_edge_ids_in_layer = go.convertTup2LoT(sid, rid)

            found_all_edges_to_be_drawn = all(
                [
                    True if el in existing_edge_ids_in_layer
                    else False
                    for el in edges_to_be_drawn
                ]
            )
            if not found_all_edges_to_be_drawn:
                raise Exception(
                    "ERROR! at least 1 of the edges between "
                    "actors/products to be drawn do NOT contain "
                    "the attribute 'weight'! please check edge "
                    "ids of corresponding nodes!"
                )

        pos = dict()
        # handle each stage sorted by actors
        all_stage_coords = []
        all_stage_names = []
        for st_id, stage in enumerate(stagewise_actors):
            # syntax : (node_id, (x-coord, y-coord)
            # x-coord = for each node ordered by node id
            # y-coord = stage

            name_of_stage = ''.join(
                list(
                    map(str, set([j for (i, j) in stage]))
                )
            )
            my_stage_coord = nr_of_stages - st_id
            pos.update(
                (nid, (i, my_stage_coord)) for i, (nid, _) in enumerate(stage)
            )

            # log all possible coordinates for later use, when drawing
            all_stage_coords.append(my_stage_coord)
            all_stage_names.append(name_of_stage)

        return pos, all_stage_coords, all_stage_names

    def node_size(self, node_attr=['output', 'input']):
        """
        purpose: node size should depend on output/throughput/input of nodes

        :param node_attr: (optional, default: ['output','input'])
        node attribute for determining node_size. needs to be defined on all
        nodes to be drawn. can be 1 attribute(str) or multiple (list of str).

        :return: node options for drawing
        """

        if isinstance(node_attr, list):
            do_they_exist = [
                attr for attr in node_attr
                if attr in go.getExistingAttrs(
                    self.SCGraph,
                    gtype='nodes'
                )
            ]
            if not do_they_exist:
                raise Exception(
                    "no attribute by the names: " +
                    str(node_attr) +
                    " exist on the graph!"
                )
            elif len(do_they_exist) < len(node_attr):
                warnings.warn(
                    "only " +
                    str(do_they_exist) +
                    " of " +
                    str(node_attr) +
                    " are attributes of the graph!"
                )
            node_list = []
            node_size = []
            for attr in node_attr:
                nodes, content = zip(
                    *nx.get_node_attributes(self.SCGraph, attr).items()
                )
                node_list = node_list + list(nodes)
                node_size = node_size + list(content)
        elif isinstance(node_attr, str):
            node_list, node_size = zip(
                *nx.get_node_attributes(self.SCGraph, node_attr).items()
            )

        return node_list, node_size

    def edge_color(self, weight='weight', color_map=plt.cm.plasma):
        """
        purpose: edge color is defined by minimum and maximum of amount of
        wares transported (std: weight)

        :param weight: (optional, default: 'weight'): edge attribute for
        coloring edges via heatmap
        :param color_map: (optional, default: plt.cm.coolwarm) blue =
        minimum, red = maximum, change only if matplotlibs colormap are
        understood!

        :return: colorbar_data, option_data for drawing
        """
        # check if weight exists on graph
        if not weight in go.getExistingAttrs(self.SCGraph, gtype='edges'):
            raise Exception(
                "given edge attribute: " +
                str(weight) +
                " does not exist in the graph"
            )

        # unzip edge ids and their respective content
        edges, weights = zip(*nx.get_edge_attributes(
            self.SCGraph, weight
        ).items())

        # for defining boundaries of color map
        min_weights = min(weights)
        max_weights = max(weights)

        edge_options = {
            "edgelist": edges,
            "edge_color": weights,
            "width": 4,
            "edge_cmap": color_map,
            "with_labels": False,
        }

        sm = plt.cm.ScalarMappable(
            cmap=color_map,
            norm=plt.Normalize(
                vmin=min_weights, vmax=max_weights
            )
        )
        sm._A = []

        return sm, edge_options



from supplychainmodelhelper import datahandling as dh
import h5py # 2.10.0
import csv
import pandas as pd # 0.25.2
import os.path
import unittest as ut


def test_decBSList():
    assert dh.decBSList([b'string1',b'string2']) == ["string1","string2"]
    assert dh.decBSList(['string1','string2']) == ["string1","string2"]

def test_decBS():
    assert dh.decBS(b'string1') == "string1"
    assert dh.decBS('string1') == "string1"



# Testing datahandling
#def test_importDataFromFolder():
#    assert dh.importDataFromFolder("./data/metadata/","folderMD.csv","./data/rawData.hdf5") == True




'''
NEED TO LEARN MORE ABOUT CLASSES... !


purpose: create graph from specifications of the user with 
generally used node properties used in supply chain models



Methods:
----------
directly operating on network
* createGraph - the graph is created with node attributes
    'actors','locations','products'
* getListOfNodeIDs - returns a list of nodeIDs with the corresponding
    attributes, used for editing information on the graph
* getNodeIDswAttr - returns a tuple of (nodeIDs,attributes) with the
    corresponding attributes, used for editing information on the graph
* addAttr2ExistingNodes - adds a user specified attribute to the nodes
    of the network
* addAttr2Edges - adds a user specified attribute to the edges of the network
* getEdgeID - returns the IDs for the edges on the graph, given the user
    specified nodes (IDs given by getListOfNodeIDs) on the network
* getEdgesAttr - returns a list of attributes of the edges
    (IDs given by getEdgeID) specified by the user
* getEdgeIDswAttr - returns a tuple of
    ((edge ID, attribute) = (node out,node in, attribute))
    of the edges (IDs given by getEdgeID) specified by the user
* getExistingAttrs - returns a list of existing attributes on the graph
    (either nodes or edges)

toolkit for handling attributes on the network
* combineActorBrand - if an actor consists of different subtypes
    (e.g. warehouse and brand(ALDI,REWE, etc.)), then this function helps
    with a fast way of creating additional actors, before creating the network
* convertLoT2NTs - functions like getEdgeIDswAttr, return a list of
    tuples, but only a part of this is needed. This function accepts a list
    of tuples as input and returns 2 lists with the same order as before.
* convertTup2LoT - functions like addAttr2ExistingNodes, accept a list
    of tuples as input, but only a list available. This function accepts
    2 lists as input and returns a list on tuples with the same order as
    before.
* getAllCombinations - getEdgeID is a function that has the corresponding
    nodes as input, and returns a direct  1-to-1 list (node out, node in).
    If all possible combinations of edges between 2 nodes are to be
    returned, this function helps with that.
* convertMatrix - 2 applications:
    1. for calibration might be needed if calibration measure is
    of different dimension as the calculated flows
    2. if a matrix with redundant information needs to be created
    from a matrix with minimum information (e.g. similar geographical
    distances between nodes to be put into a distance matrix)

mapping raw data objects to network data objects
* proxyModel - simple fraction model. If data for correlated variable
    is available, this model may be used to use the proxy data to model
    the data in question
    (e.g. modellled data = produced eggs, proxy data= number of laying hens)
* getDistMatrix - accesses the graph for attribute 'distance' for
    user-specified nodes, return a dataframe containing information of the
    geographical distance between the locations of given nodes (needs to be
    entered into the graphe beforehand via the function addAttr2Edges)
    NOTE that the attribute 'distance' needs to be defined on the edges
    of the participating nodes.
* optFlow, minOptFlowForBeta - calculating and minimizing the optimal
    flow given supply and demand of a product and the distance between
    participating nodes. Needed for calculating minimum transport
    distance (see calcPotTransportDistances).
    NOTE that the attribute 'distance' needs to be defined on the edges
        of the participating nodes.
    NOTE that the attribute 'output' needs to be defined on the nodes
        supplying the product.
    NOTE that the attribute 'distance' needs to be defined on the edges
        of the participating nodes.
* propFlow - calculating the proportional flow given supply and demand
    of a product and the distance between participating nodes. Needed for
    calculating maximum transport distance (see calcPotTransportDistances).
    NOTE that the attribute 'distance' needs to be defined on the edges
        of the participating nodes.
    NOTE that the attribute 'output' needs to be defined on the nodes
        supplying the product.
    NOTE that the attribute 'distance' needs to be defined on the edges
        of the participating nodes.
* furnessModel - given geographical distance, supply and demand of
    the participating nodes and the free parameter (resistance factor)
    beta this function returns the flow between given nodes
    (beta needs to be determined by hymanModel / calibration)
    NOTE that the attribute 'distance' needs to be defined on the edges
        of the participating nodes.
    NOTE that the attribute 'output' needs to be defined on the nodes
        supplying the product.
    NOTE that the attribute 'input' needs to be defined on the nodes
        demanding the product.
* calcPotTransportDistances - calculates the range of possible average
    transport distances and returns a list of values within this interval.
    The number of returned values from this interval may be specified by
    the user (needed for hymanModel to determine beta for furnessModel).
    NOTE that each of the returned values are POSSIBLE average
    transport distance. The best one still needs to be determined(see
    calibration).
    NOTE that the attribute 'distance' needs to be defined on the
        edges of the participating nodes.
    NOTE that the attribute 'output' needs to be defined on the nodes
        supplying the product.
    NOTE that the attribute 'input' needs to be defined on the nodes
        demanding the product.
* hymanModel - calculates resistance factor beta given a transport
    distance. Returns the flow between participating nodes of supply and
    demand.
    NOTE that each of the returned values are POSSIBLE average transport
        distance. The best one still needs to be determined(see calibration).
    NOTE that the attribute 'distance' needs to be defined on the edges
        of the participating nodes.
    NOTE that the attribute 'output' needs to be defined on the nodes
        supplying the product.
    NOTE that the attribute 'input' needs to be defined on the nodes
        demanding the product.
* getWeightedDist - returns the average transport distance, given the flow
    between participating nodes of supply and demand.
    NOTE that each of the returned values are POSSIBLE average
        transport distance. The best one still needs to be determined(see
        calibration).
    NOTE that the attribute 'distance' needs to be defined on the edges
        of the participating nodes.
    NOTE that the attribute 'output' needs to be defined on the nodes
        supplying the product.
    NOTE that the attribute 'input' needs to be defined on the nodes
        demanding the product.
* calibration - function for calculating the best configuration of given
    transport distances and corresponding flows. A number of possible
    average transport distances are given by the function
    calcPotTransportDistances. The function hymanModel calculates
    the corresponding flows.
    Now this needs to be calibrated against a measure to determine which
    average transport distance fits best the calibration measure.
    This function accepts a 4D-Tensor of all flows as input
    [supply,demand,product,transport distance] and the calibration measure
    in question.
    NOTE that each of the returned values are POSSIBLE average
        transport distance. The best one still needs to be determined(see
        calibration).
    NOTE that the attribute 'distance' needs to be defined on the edges
        of the participating nodes.
    NOTE that the attribute 'output' needs to be defined on the nodes
        supplying the product.
    NOTE that the attribute 'input' needs to be defined on the nodes
        demanding the product.

* calibrationOLD - deprecated function for calculating the best
    configuration of given transport distances and corresponding flows.
    MCMC simulation. Only here for reasons of nostalgia.




===============================================================================
'''

import networkx as nx
from networkx.classes.digraph import DiGraph
#from networkx.classes.graph import Graph
import numpy as np
from itertools import chain
from ipfn import ipfn
from scipy import optimize
import warnings
warnings.filterwarnings("ignore")

class SCgraph(DiGraph):
    '''
    purpose: create graph from specifications of the user with 
    generally used node properties used in supply chain models



    Methods:
    ----------
    directly operating on network
    * createGraph - the graph is created with node attributes
    'actors','locations','products'
    * getListOfNodeIDs - returns a list of nodeIDs with the corresponding
        attributes, used for editing information on the graph
    * getNodeIDswAttr - returns a tuple of (nodeIDs,attributes) with the
        corresponding attributes, used for editing information on the graph
    * addAttr2ExistingNodes - adds a user specified attribute to the nodes
        of the network
    * addAttr2Edges - adds a user specified attribute to the edges of the
        network
    * getEdgeID - returns the IDs for the edges on the graph, given the
        user specified nodes (IDs given by getListOfNodeIDs) on the network
    * getEdgesAttr - returns a list of attributes of the edges
        (IDs given by getEdgeID) specified by the user
    * getEdgeIDswAttr - returns a tuple of
        ((edge ID, attribute) = (node out,node in, attribute))
        of the edges (IDs given by getEdgeID) specified by the user
    * getExistingAttrs - returns a list of existing attributes on the
        graph (either nodes or edges)

    toolkit for handling attributes on the network
    * combineActorBrand - if an actor consists of different subtypes
        (e.g. warehouse and brand(ALDI, REWE, etc.)), then this function
        helps with a fast way of creating additional actors, before creating
        the network
    * convertLoT2NTs - functions like getEdgeIDswAttr, return a list of
        tuples, but only a part of this is needed. This function accepts
        a list of tuples as input and returns 2 lists with the same order as
        before.
    * convertTup2LoT - functions like addAttr2ExistingNodes, accept a list of
        tuples as input, but only a list available. This function accepts 2
        lists as input and returns a list on tuples with the same order as
        before.
    * getAllCombinations - getEdgeID is a function that has the corresponding
        nodes as input, and returns a direct  1-to-1 list (node out, node in).
        If all possible combinations of edges between 2 nodes are to be
        returned, this function helps with that.
    * convertMatrix - 2 applications:
        1. for calibration might be needed if calibration measure is of
            different dimension as the calculated flows
        2. if a matrix with redundant information needs to be created from
            a matrix with minimum information (e.g. similar geographical
            distances between nodes to be put into a distance matrix)

    mapping raw data objects to network data objects
    * proxyModel - simple fraction model. If data for correlated variable
        is available, this model may be used to use the proxy data to model
        the data in question (e.g. modellled data = produced eggs,
        proxy data= number of laying hens)
    * getDistMatrix - accesses the graph for attribute 'distance'
        for user-specified nodes, return a dataframe containing information
        of the geographical distance between the locations of given nodes
        (needs to be entered into the graphe beforehand via the
        function addAttr2Edges)
        NOTE that the attribute 'distance' needs to be defined on the edges
            of the participating nodes.
    * optFlow, minOptFlowForBeta - calculating and minimizing the optimal
        flow given supply and demand of a product and the distance between
        participating nodes. Needed for calculating minimum transport
        distance (see calcPotTransportDistances).
        NOTE that the attribute 'distance' needs to be defined on the edges
            of the participating nodes.
        NOTE that the attribute 'output' needs to be defined on the nodes
            supplying the product.
        NOTE that the attribute 'distance' needs to be defined on the edges
            of the participating nodes.
    * propFlow - calculating the proportional flow given supply and demand
        of a product and the distance between participating nodes. Needed
        for calculating maximum transport distance
        (see calcPotTransportDistances).
        NOTE that the attribute 'distance' needs to be defined on the
            edges of the participating nodes.
        NOTE that the attribute 'output' needs to be defined on the
            nodes supplying the product.
        NOTE that the attribute 'distance' needs to be defined on the
            edges of the participating nodes.
    * furnessModel - given geographical distance, supply and demand of the
        participating nodes and the free parameter (resistance factor)
        beta this function returns the flow between given nodes
        (beta needs to be determined by hymanModel / calibration)
        NOTE that the attribute 'distance' needs to be defined on the
            edges of the participating nodes.
        NOTE that the attribute 'output' needs to be defined on the
            nodes supplying the product.
        NOTE that the attribute 'input' needs to be defined on the
            nodes demanding the product.
    * calcPotTransportDistances - calculates the range of possible
        average transport distances and returns a list of values within
        this interval. The number of returned values from this interval may
        be specified by the user (needed for hymanModel to determine
        beta for furnessModel).
        NOTE that each of the returned values are POSSIBLE average
        transport distance. The best one still needs to be determined
        (see calibration).
        NOTE that the attribute 'distance' needs to be defined on the edges
            of the participating nodes.
        NOTE that the attribute 'output' needs to be defined on the nodes
            supplying the product.
        NOTE that the attribute 'input' needs to be defined on the nodes
            demanding the product.
    * hymanModel - calculates resistance factor beta given a transport
        distance. Returns the flow between participating nodes of supply
        and demand.
        NOTE that each of the returned values are POSSIBLE average transport
            distance. The best one still needs to be determined(see calibration).
        NOTE that the attribute 'distance' needs to be defined on the edges
            of the participating nodes.
        NOTE that the attribute 'output' needs to be defined on the nodes
            supplying the product.
        NOTE that the attribute 'input' needs to be defined on the nodes
            demanding the product.
    * getWeightedDist - returns the average transport distance, given the
        flow between participating nodes of supply and demand.
        NOTE that each of the returned values are POSSIBLE average
            transport distance. The best one still needs to be determined
            (see calibration).
        NOTE that the attribute 'distance' needs to be defined on the edges
            of the participating nodes.
        NOTE that the attribute 'output' needs to be defined on the nodes
            supplying the product.
        NOTE that the attribute 'input' needs to be defined on the nodes
            demanding the product.
    * calibration - function for calculating the best configuration of given
        transport distances and corresponding flows. A number of possible
        average transport distances are given by the function
        calcPotTransportDistances. The function hymanModel calculates
        the corresponding flows.
        Now this needs to be calibrated against a measure to determine
        which average transport distance fits best the calibration measure.
        This function accepts a 4D-Tensor of all flows as input
        [supply,demand,product,transport distance]
        and the calibration measure in question.
        NOTE that each of the returned values are POSSIBLE average
        transport distance. The best one still needs to be determined(see
        calibration).
        NOTE that the attribute 'distance' needs to be defined on the edges
            of the participating nodes.
        NOTE that the attribute 'output' needs to be defined on the nodes
            supplying the product.
        NOTE that the attribute 'input' needs to be defined on the nodes
            demanding the product.

    * calibrationOLD - deprecated function for calculating the best
        configuration of given transport distances and corresponding flows.
        MCMC simulation. Only here for reasons of nostalgia.
    '''
    
    myAttributes = ['actors', 'locations', 'products']

    def __init__(
            self,
            actors: list,
            locations: list,
            products: list,
            *args,
            **kwargs
    ):
        """
        purpose: create a directed graph with standard attributes for supply chain modelling:
        on creation of the graph a number of nodes are created based on the input lists.
        The nodes contain all possible combinations of the elements from the input list.
        The attributes are: 'actor','location','product'

        :param actors: a list of participating actors (list of string)
        :param locations: a list of locations involved (e.g. Landkreise in Germany)
        :param products: a list of products being transported (e.g. vegetables, eggs, ...)

        :return: Graph with nodes containing all possible combinations of actors, locations
        and products

        example:
        from supplychainmodulator import graphoperations as go

        prod = ['milk','beer','schnaps']
        act = ['producer','consumer','warehouse','store']
        loc = ['BER','SXF','TXL']

        myNW = go.createGraph(listOfActors=act,listOfLocations=loc,\\
            listOfProducts=prod)




        =======================================================================
        """
        super(SCgraph, self).__init__(*args, **kwargs)
        self.actors = actors
        self.products = products
        self.locations = locations

        isItListA = isinstance(self.actors, list)
        isItListP = isinstance(self.products, list)
        isItListL = isinstance(self.locations, list)
        
        if not isItListA:
            raise Exception('actors ist not a list!')
        if not isItListP:
            raise Exception('products ist not a list!')
        if not isItListL:
            raise Exception('locations ist not a list!')

        howManyActors = len(self.actors)
        howManyProducts = len(self.products)
        howManyLocations = len(self.locations)
    
        if howManyActors == 0:
            raise Exception(
                'incorrect parameter: '
                'actors should be a list, but it is: %20s' % (actors)
            )
        if howManyProducts == 0:
            raise Exception(
                'incorrect parameter: '
                'products should be a list, but it is: %20s' % (products)
            )
        if howManyLocations == 0:
            raise Exception(
                'incorrect parameter: '
                'locations should be a list, but it is: %20s' % (locations)
            )

        
        

        # create names for node attributes
        # (currently hardcoded, since this never changes (as of this moment))
        actor = self.myAttributes[0]
        location = self.myAttributes[1]
        product = self.myAttributes[2]

        '''
        go through all permutations and create nodes with 
        these 3 attribute names and a list of the 
        corresponding 3 attribute contents
        '''
        nodes = zip(
            range(
                len(self.actors)*len(self.locations)*len(self.products)
            ), [
                {actor:al,product:pl,location:ll}
                for al in self.actors
                for pl in self.products
                for ll in self.locations
            ]
        )
        
        self.add_nodes_from(list(nodes))

    def combineActorBrand(
            self,
            actor: str,
            brand: list
    ) -> list:
        """
        purpose: if an actor has a brand, but is the same type, this can be used
        to give it the correct designation
        (e.g. retailer: Warehouse_ALDI or Store_ALDI a.s.o.)
        usually such an actor type has multiple brands

        :param actor: an element of the actor list
        :param brand: a list of brands for different actors: actor_suffix

        :return: a combined list

        example:
        from supplychainmodelhelper import graphoperations as go

        act = ['producer','consumer','warehouse','store']

        newActors = go.combineActorBrand(actor='warehouse',brand=['ALDI','REWE','LIDL'])

        act.remove('warehouse')
        act.extend(newActors)

        print(act)


        ===============================================================================================
        """
        # TODO: sanity check if given actor is in graph
        # TODO: remove actor from graph, add new actors to graph
    
        combinedList = [actor + "_" + go for go in brand]
        return combinedList


# import csv #1.0
import h5py  # 2.10.0
import pandas as pd  # 0.25.2
import os.path
# import sys
# sys.path.append('../')
from supplychainmodelhelper.datahandling import decBSList
import warnings


# warnings.filterwarnings("ignore")

class Datacube:
    '''
    purpose: easy editing of hdf5 database for supply chain modelling
    consists of basic meta data schema for database itself, all 
    folders and datasets.
    basic structure
    database
        - metadata information of all containing folders
        - folder
            - metadata information about folder
            - dataset
                - metadata information about dataset
                - data of dataset
                - axis of dataset

    Possible actions:
    - initialise database: a new hdf5 file created or an existing accessed 
    (see __init__)
    
    - extend the existing basic metadata schema of the database
    Please NOTE that only 1 schema exists at a time for the
    database in question. This schema exists to get easy
    access to existing folders and scan through the database rather
    quickly.
    At this point there is no plan to add multiple md schemas
    to 1 database.
    (add2TemplateMDofDB)
    
    - extend the existing basic metadata schema of an existing dataset
    Please NOTE that only 1 schema exists at a time for the
    dataset in question. This schema exists to get easy
    access to existing datasets and scan through the database rather
    quickly.
    At this point there is no plan to add multiple md schemas
    to 1 database.
    Further note that folders have a similar md schema as the dataset
    itself. The difference is that in any folder the md schema for
    ALL datasets inside the folder is summarized. See getMDFromFolder
    and getMDFromDataSet.
    (add2TemplateMDofDataset)
    
    - store the current metadata schema of the database as a 
    template csv file, for later import (createTemplate_MD_Schema_DB_CSV)
    
    - store the current metadata schema of a folder as a 
    template csv file, for later import (createTemplate_MD_Schema_Folder_CSV)
    
    - import csv file with current metadata schema of folder
    and filled out metadata information about containing datasets 
    (importFromCSV_MD_DataForFolder)
    
    -  import csv file with current metadata schema of database
    and filled out metadata information about containing datasets
    (importFromCSV_MD_DataForDB)
        
    - add/remove a folder to the database, incl. a list of metadata information
    based on the current metadataschema (addFolder2ExistingDB, removeFolder)
    
    - add/remove a dataset to an existing folder, incl. a list of metadata
    information based on the current metadataschema
    (addDataSet2ExistingFolder, removeDatasetFromFolder)
    
    - get an existing dataset from a specific folder in the database
    (getDataFrameFromFolder)
    
    - get metadata information about an existing dataset in the database
    (getMDFromDataSet)
    
    - get metadata information about an existing folder in the database
    (getMDFromFolder)
    
    - get metadata information about the database
    (getMDfrom_db)

    - change metadata information of dataset, folder, database
    (editMDofDB,editMDofDataset)

    - remove meta data categories from md schema
    (removeFolderMD,removeDatasetMD)

    '''
    # basic template for mandatory entries for metadata schema
    # if needed add entries in these lists
    # with add2TemplateMDofDB or add2TemplateMDofDataset
    h5file_name = 'defaultDB.hdf5'
    listOfTemplateMDofDB = [
        'Folder',
        'ID',
        'Title',
        'Description'
    ]
    list_template_md_dataset = [
        'Dataset',
        'file_name',
        'Format',
        'ID',
        'Title',
        'Rows',
        'Columns',
        'Encoding'
    ]

    # TODO test in combination with other functions
    def __init__(
            self,
            h5file_name,
            rights='add'
    ):
        '''
        purpose: initialise the data object

        input:
        :param h5file_name(optional): path to the hdf5 file. If not existing, a
        new file will be created. default file_name is 'defaultDB.hdf5'
        :param rights(optional): user rights to access this database. 
        'add'(default) user may read/write, if no file by this name exists, the 
        file will be created.
        'new' will overwrite any existing databases (good for testing the
        database, less good for working with it).
        Also on initialisation of this data object the given hdf5 database
        will be accessed and the md schema from the db overwrites the
        default md schema

        :return: none
        
        example:
        from supplychainmodelhelper import mds
        myNewDB = mds.Datacube('test.hdf5','new')

        if needed the file_name is attached to this data object
        print(testDC.h5file_name)



        ======================================================================


        '''

        self.overwritefile_name(h5file_name)
        if rights == 'add':
            with h5py.File(h5file_name, 'a') as db:
                # check md schema of existing db
                if len(list(db.keys())) > 0:
                    check_current_md_schema = list(self.getMDfrom_db().columns)
                    all_categories = [
                        True if i in self.listOfTemplateMDofDB
                        else False
                        for i in check_current_md_schema
                    ]
                    # check if more categories exist in db than in basic md
                    # schema
                    if not all(all_categories):
                        # get index of new categories
                        add_cat = [
                            f for f, x in enumerate(all_categories)
                            if not x
                        ]
                        self.listOfTemplateMDofDB += list(
                            map(
                                check_current_md_schema.__getitem__,
                                add_cat
                            )
                        )
                    new_categories = []
                    for folder in db.keys():
                        for dataset in db[folder].keys():
                            check_current_md_schema = list(
                                self.getMDFromDataSet(
                                    folder,
                                    dataset
                                ).columns
                            )
                            all_categories = [
                                True if i in self.listOfTemplateMDofDB
                                else False
                                for i in check_current_md_schema
                            ]
                            if not all(all_categories):
                                add_cat = [
                                    f for f, x in enumerate(all_categories)
                                    if not x
                                ]
                                new_cat_this_time = list(
                                    map(
                                        check_current_md_schema.__getitem__,
                                        add_cat
                                    )
                                )
                                new_categories = new_categories + \
                                                 new_cat_this_time
                    if new_categories != new_cat_this_time:
                        warnings.warn(
                            'Please be advised: different meta data schemas '
                            'for different datasets is untested!'
                        )
                    self.list_template_md_dataset += new_categories
        elif rights == 'new':
            with h5py.File(self.h5file_name, 'w'):
                print('new file created...' + self.h5file_name)
        else:
            raise Exception('please choose either \'add\' or \'new\'!')

    @classmethod
    def overwritefile_name(
            cls,
            file_name
    ):
        cls.h5file_name = file_name

    # TODO test in combination with other functions
    # TODO write self test to test new db sync functionality
    @classmethod
    def add2TemplateMDofDB(
            cls,
            list_additional_categories,
            sure=False
    ):
        '''

        if folder/dataset in db exists, add this new element to all existing
        give user warning that missing entries need to be added to dataset by
        other function
        purpose: adding a list of metadata categories to the overall metadata
        structure of the database

        input:
        :param list_additional_categories: a list of new categories
        :param sure(optional) only asked for if database named in
        initialisation, already contains folders.
        New categories in db will have the note 'no entry yet'


        :return: none

        example:
        from supplychainmodelhelper import mds

        # initialising the datacube operations toolkit
        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname,'new')

        # basic md template on creation
        print(testDC.listOfTemplateMDofDB)
        
        # extend the metadata schema for database
        testDC.add2TemplateMDofDB(['db category 1','db category 2',
        'db category 3'])
        
        # check out the CURRENT metadata schema of the database
        print(testDC.listOfTemplateMDofDB)
 


        =======================================================================
        '''
        # new md schema is built
        new_md_schema = cls.listOfTemplateMDofDB[:] + \
                        list_additional_categories[:]

        # check if existing folders need update
        new_cat = [
            i for i in new_md_schema
            if i not in cls.listOfTemplateMDofDB
        ]

        # check with db
        with h5py.File(cls.h5file_name, 'a') as db:
            # do folders exist? if so, check if element exists and
            # if not -> update it
            if not sure:
                if len(list(db.keys())) > 0:
                    warnings.warn(
                        'Please be advised. '
                        '\nYou are about to change the md schema '
                        'on existing db. \nRepeat this command '
                        'with the flag sure=True, '
                        'if you\'re really sure about this'
                    )
                else:
                    # counting doubles
                    if len(new_md_schema) == len(set(new_md_schema)):
                        cls.listOfTemplateMDofDB = new_md_schema[:]
                    else:
                        warnings.warn(
                            'At least one element was already '
                            'in the md schema. '
                            '\nRemoving the redundancy!'
                        )
                        cls.listOfTemplateMDofDB = list(set(new_md_schema[:]))
            # user is sure about what he/she is doing
            if sure:
                if len(list(db.keys())) > 0:
                    # go through every folder
                    for run in new_cat:
                        # for each folder 1 entry exists
                        # if this is the first entry
                        if not db.attrs.__contains__(run):
                            db.attrs[run] = ['no entry yet'] * len(new_cat)
                    # counting doubles
                    if len(new_md_schema) == len(set(new_md_schema)):
                        cls.listOfTemplateMDofDB = new_md_schema[:]
                    else:
                        warnings.warn(
                            'Warning: At least one element was already'
                            ' in the md schema, removing the redundancy.'
                        )
                        cls.listOfTemplateMDofDB = list(set(new_md_schema[:]))
                else:
                    warnings.warn(
                        'Warning: You set the flag \'sure\'=True, '
                        'but no entries are to be find in database! '
                        'Nothing done.'
                    )

    # TODO test in combination with other functions
    # TODO write self test to test new db sync functionality
    @classmethod
    def add2TemplateMDofDataset(
            cls,
            list_additional_categories,
            sure=False
    ):
        '''
        purpose: adding a list of metadata categories to the folder meta data

        input:
        :param list_additional_categories: a list of new categories 
        :param sure(optional) only asked for if an hdf5 contains folder or
        dataset and the needs to be updated afterwards. New categories in db
        will have the note 'no entry yet'

        :return: none

        example:
        from supplychainmodelhelper import mds

        # initialising the datacube operations toolkit
        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname,'new')

        # basic md template on creation
        print(testDC.list_template_md_dataset)
        
        # extend the metadata schema for database
        testDC.add2TemplateMDofDataset(['data category 1','data category 2',
        'data category 3'])
        
        # check out the CURRENT metadata schema of the database
        print(testDC.list_template_md_dataset)

        # if dataset are added, the flag 'sure' needs to be set to 'True'
        otherwise change requests are ignored
        testDC.addFolder2ExistingDB(['thisnewFolder', '2', 's',
        'Descripcvdxction'])

        =======================================================================
        
        '''
        new_md_schema = cls.list_template_md_dataset + \
                      list_additional_categories

        # check if existing folders need update
        new_cat = [
            i for i in new_md_schema
            if i not in cls.list_template_md_dataset
        ]

        # check with db
        with h5py.File(cls.h5file_name, 'a') as db:
            # do folders exist? if so, check if element exists and
            # if not -> update it
            if not sure:
                if len(list(db.keys())) > 0:
                    warnings.warn(
                        'Please be advised. You are about to '
                        'change the md schema on existing db.'
                        '\nRepeat this command with the flag '
                        'sure=True, if you\'re really sure about '
                        'this'
                    )
                else:
                    # counting doubles
                    if len(new_md_schema) == len(set(new_md_schema)):
                        cls.list_template_md_dataset = new_md_schema
                    else:
                        warnings.warn(
                            'at least one element was already in the md '
                            'schema, removing the redundancy.'
                        )
                        cls.list_template_md_dataset = list(set(new_md_schema))

            # user is sure about what he/she is doing
            if sure:
                if len(list(db.keys())) > 0:
                    # go through every folder
                    for folder in db.keys():
                        for run in new_cat:
                            # for each folder 1 entry exists
                            # if this is the first entry
                            if not db[folder].attrs.__contains__(run):
                                db[folder].attrs[run] = 'no entry yet'
                        for ds in db[folder].keys():
                            for run in new_cat:
                                if not db[
                                    folder + '/' + ds
                                ].attrs.__contains__(run):
                                    db[
                                        folder + '/' + ds
                                    ].attrs[run] = 'no entry yet'
                    # counting doubles
                    if len(new_md_schema) == len(set(new_md_schema)):
                        cls.list_template_md_dataset = new_md_schema
                    else:
                        warnings.warn(
                            'at least one element was already in the '
                            'md schema. '
                            '\nRemoving the redundancy.'
                        )
                        cls.list_template_md_dataset = list(set(new_md_schema))
                else:
                    warnings.warn(
                        'Warning: You set the flag \'sure\'=True, but no '
                        'entries are to be find in database! Nothing done.'
                    )

    # TODO test in combination with other functions
    # TODO write self test to test new db sync functionality
    # TODO some elements (name, id) cant be removed
    @classmethod
    def removeFromTemplateMDofDB(
            cls,
            attr2remove,
            sure=False
    ):
        '''
        purpose: 
        If no folders are inserted yet to the database, the md schema
        wil be updated with no further warning. This can be undone
        by the function add2TemplateMDofDB.
        If at least 1 folder in db exists, remove this category from all
        existing folders in db. The user gets a warning if the flag 'sure'
        is NOT set to true.

        input:
        :param attr2remove: can be a list or a str
        :param sure(optional): set True, if database exists already and
        existing entries are to be removed

        :return: none

        example:
        from supplychainmodelhelper import mds

        # initialising the datacube operations toolkit
        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname,'new')

        # basic md template on creation
        print(testDC.listOfTemplateMDofDB)

        # extend the metadata schema for database
        testDC.add2TemplateMDofDB(
            ['db category 1','db category 2','db category 3']
        )

        # check out the CURRENT metadata schema of the database
        print(testDC.listOfTemplateMDofDB)

        # extend the metadata schema for database
        # can be a list
        testDC.removeFromTemplateMDofDB(['db category 1','db category 2'])
        # or a string
        testDC.removeFromTemplateMDofDB('db category 3')

        # check out the CURRENT metadata schema of the database
        print(testDC.listOfTemplateMDofDB)


        =======================================================================
        '''
        if not sure:
            with h5py.File(cls.h5file_name, 'r') as db:
                # do folders exist?
                # if so, check if element exists and if not update it
                if len(list(db.keys())) > 0:
                    # dont do anything if folders exists
                    # and the flag sure is not set
                    # -> probably user did something wrong
                    rem = False

                    if isinstance(attr2remove, str):
                        if db.attrs.__contains__(attr2remove):
                            warnings.warn(
                                'Please be advised. ' +
                                str(attr2remove) +
                                ' is already entered into the database.\n'
                                'Repeat this command with the flag sure=True,'
                                ' if you\'re really sure about this'
                            )
                        else:
                            warnings.warn(
                                'Attribute '
                                + str(attr2remove) +
                                ' not found in database. '
                                'Please check spelling!'
                            )
                    elif isinstance(attr2remove, list):
                        for el in attr2remove:
                            if db.attrs.__contains__(el):
                                warnings.warn(
                                    'Please be advised. ' +
                                    str(el) +
                                    ' is already entered into the database.\n'
                                    'Repeat this command with the flag '
                                    'sure=True, if you\'re really sure about '
                                    'this'
                                )
                            else:
                                warnings.warn(
                                    'Attribute ' +
                                    str(el) +
                                    ' not found in database. '
                                    'Please check spelling!'
                                )
                else:
                    # no folders exist and sure is not set to True
                    # by hand of user -> probably ok,
                    # but if not, user may undo this by add2TemplateMDofDB
                    rem = True
        if sure:
            with h5py.File(cls.h5file_name, 'a') as db:
                # user is sure about what he/she is doing
                if len(list(db.keys())) > 0:
                    # probably ok, so go ahead and update class variable
                    # with md schema
                    rem = True
                    if isinstance(attr2remove, str):
                        if db.attrs.__contains__(attr2remove):
                            del db.attrs[attr2remove]
                    elif isinstance(attr2remove, list):
                        for el in attr2remove:
                            if db.attrs.__contains__(el):
                                del db.attrs[el]
                else:
                    # probably the user did something wrong,
                    # so dont do anythind
                    rem = False
                    warnings.warn(
                        'You set the flag sure, even if no database could be '
                        'found. Either read documentation more carefully, or '
                        'something weird happened. For security, nothing '
                        'removed.'
                    )

        # rem is true if sure is False AND no existing entries are in db
        # -> entries removed
        # rem is false if sure is True AND no existing entries are in db
        # -> something weird happened or user didnt read the doc
        # rem is false if sure is False AND existing entries are in db
        # -> user tries to edit db without being sure
        # rem is true if sure is True AND existing entries are in db
        # -> entries removed and synced with db
        if rem:
            # check if input is a list or a str
            if isinstance(attr2remove, str):
                if attr2remove not in cls.listOfTemplateMDofDB:
                    raise Exception(
                        'Given string is not an element of '
                        'the current meta data schema!'
                    )
                else:
                    cls.listOfTemplateMDofDB.remove(attr2remove)
            elif isinstance(attr2remove, list):
                for a in attr2remove:
                    if a not in cls.listOfTemplateMDofDB:
                        raise Exception(
                            'Element ' +
                            str(a) +
                            ' is not an element of the '
                            'current meta data schema!'
                        )
                    cls.listOfTemplateMDofDB.remove(a)
            else:
                raise Exception(
                    'Input is neither list nor string! Nothing removed!'
                )

    # TODO test in combination with other functions
    # TODO write self test to test new db sync functionality
    # TODO some elements (name, id) cant be removed
    @classmethod
    def removeElementTemplateDataSetMD(
            cls,
            attr2remove,
            sure=False
    ):
        """
        purpose: removing either 1 element as string or multiple elements of
        metadata categories from ALL datasets.
        There are no multiple schemas for different datasets.
        Might be added later, if pressure is high.

        input:
        :param attr2remove: a str or a list of existing categories
        :param sure: (Boolean) only set to true, if you're sure
        :return: none

        example:
        from supplychainmodelhelper import mds

        # initialising the datacube operations toolkit
        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname,'new')

        # basic md template on creation
        print(testDC.list_template_md_dataset)

        # extend the metadata schema for database
        testDC.add2TemplateMDofDataset(
            ['folder category 1','folder category 2','folder category 3']
        )

        # check out the CURRENT metadata schema of the database
        print(testDC.list_template_md_dataset)

        # removing 1 item
        testDC.removeElementTemplateDataSetMD('folder category 2')
        print(testDC.list_template_md_dataset)

        # removing more items
        testDC.removeElementTemplateDataSetMD(
            ['folder category 1','folder category 3']
        )
        print(testDC.list_template_md_dataset)


        =======================================================================

        """
        if not sure:
            with h5py.File(cls.h5file_name, 'r') as db:
                # do folders exist?
                # if so, check if element exists and if not update it
                if len(list(db.keys())) > 0:
                    # dont do anything
                    # if folders exists and the flag sure is not set
                    # -> probably user did something wrong
                    rem = False

                    for folder in db.keys():
                        if isinstance(attr2remove, str):
                            if db[folder].attrs.__contains__(attr2remove):
                                warnings.warn(
                                    'Please be advised. ' +
                                    str(attr2remove) +
                                    ' is already entered into the database.\n'
                                    'Repeat this command with the flag '
                                    'sure=True, if you\'re really sure '
                                    'about this')
                            else:
                                warnings.warn(
                                    'Attribute ' +
                                    str(attr2remove) +
                                    ' not found in database. '
                                    'Please check spelling!'
                                )
                        elif isinstance(attr2remove, list):
                            for el in attr2remove:
                                if db[folder].attrs.__contains__(el):
                                    warnings.warn(
                                        'Please be advised. ' +
                                        str(el) +
                                        ' is already entered into the '
                                        'database.\nRepeat this command with '
                                        'the flag sure=True, if you\'re '
                                        'really sure about this'
                                    )
                                else:
                                    warnings.warn(
                                        'Attribute ' +
                                        str(el) +
                                        ' not found in database. '
                                        'Please check spelling!'
                                    )
                else:
                    # no folders exist and sure
                    # is not set to True by hand of user
                    # -> probably ok, but if not,
                    # user may undo this by add2TemplateMDofDB
                    rem = True

        # optional flag default is False
        if sure:
            with h5py.File(cls.h5file_name, 'a') as db:
                # do folders exist?
                # if so, check if element exists and if not update it
                if len(list(db.keys())) > 0:
                    # probably ok, so go ahead and update class variable
                    # with md schema
                    rem = True

                    for folder in db.keys():
                        if isinstance(attr2remove, str):
                            if db[folder].attrs.__contains__(attr2remove):
                                del db[folder].attrs[attr2remove]
                                print(
                                    'attribute removed from folder: ' +
                                    folder
                                )
                            else:
                                warnings.warn(
                                    'Attribute ' +
                                    str(attr2remove) +
                                    ' not found in database. '
                                    'Please check spelling!'
                                )
                            for ds in db[folder].keys():
                                if db[folder + '/' + ds].attrs.__contains__(
                                        attr2remove
                                ):
                                    del db[
                                        folder + '/' + ds
                                        ].attrs[attr2remove]
                                    print(
                                        'attribute removed from dataset: ' +
                                        ds
                                    )
                        elif isinstance(attr2remove, list):
                            for el in attr2remove:
                                if db[folder].attrs.__contains__(el):
                                    del db[folder].attrs[el]
                                    print(
                                        'attribute removed from folder: ' +
                                        folder
                                    )
                                else:
                                    warnings.warn(
                                        'Attribute ' +
                                        str(el) +
                                        'in folder ' +
                                        folder +
                                        ' not found in database. '
                                        'Please check spelling!'
                                    )
                                for ds in db[folder].keys():
                                    if db[
                                        folder + '/' + ds].attrs.__contains__(
                                        el
                                    ):
                                        del db[folder + '/' + ds].attrs[el]
                                        print(
                                            'attribute removed '
                                            'from dataset: ' +
                                            ds
                                        )
                                    else:
                                        warnings.warn(
                                            'Attribute ' +
                                            str(el) +
                                            'in folder ' +
                                            folder +
                                            ' in dataset ' +
                                            ds +
                                            ' not found in database. '
                                            'Please check spelling!'
                                        )
                else:
                    # probably the user did something wrong,
                    # so dont do anythind
                    rem = False
                    warnings.warn(
                        'You set the flag sure, even if no database '
                        'could be found. Either read documentation more '
                        'carfully, or something weird happened.'
                    )

        # rem is true if sure is False AND no existing entries are in db
        # -> entries removed
        # rem is false if sure is True AND no existing entries are in db
        # -> something weird happened or user didnt read the doc
        # rem is false if sure is False AND existing entries are in db
        # -> user tries to edit db without being sure
        # rem is true if sure is True AND existing entries are in db
        # -> entries removed and synced with db
        if rem:
            if isinstance(attr2remove, str):
                if attr2remove not in cls.list_template_md_dataset:
                    raise Exception(
                        'Given string is not an element of the '
                        'current meta data schema!'
                    )
                else:
                    cls.list_template_md_dataset.remove(attr2remove)
            elif isinstance(attr2remove, list):
                for a in attr2remove:
                    if a not in cls.list_template_md_dataset:
                        raise Exception(
                            'Element ' +
                            str(a) +
                            ' is not an element of the '
                            'current meta data schema!'
                        )
                    cls.list_template_md_dataset.remove(a)
            else:
                raise Exception('input is neither list nor string')

    # TODO test
    # TODO write self test to test new db sync functionality
    # TODO check if ID is unique
    def editMDofDataset(
            self,
            folder: str,
            dataset: str,
            category: str,
            content: str
    ):
        '''
        purpose:
        adding a new category to md schema of dataset is done by
        the function add2TemplateMDofDataset (see example).
        Using this function only editing of EXISTING md categories is
        supported.
        Adding a new category means an update to all datasets, while
        this function only edits (or adds) content to 1 dataset at a time.

        This function checks if all inputs (folder, dataset, category) exists
        before editing the field.

        NOTE that the folder md data is updated as well, since the folder md
        is just a summary of what is in the folder.

        TODO:
        Future updates may have the option of having the variable dataset as a
        list and content as a list, so that multiple entries may be done at
        the same time.

        TODO:
        Is a flag for overwriting existing content necessary?

        input:
        :param folder: existing folder in db (str) - assumed all datasets are
        in given folder
        :param dataset: existing dataset (str)
        :param category: existing category (str) for changing 1 category
        at a time
        :param content: corresponding content existing category for changing

        :return: none

        example: 
        from supplychainmodelhelper import mds
        myNewDB = mds.Datacube('test.hdf5','new')

        # adding folders to database
        newFolderMD = [
            'newfolder',
            '1',
            'A new Folder',
            'A very cool description'
        ]
        myNewDB.addFolder2ExistingDB(newFolderMD)

        # adding dataset to folder
        loc = ['BER','TXL','SXF']
        my_data={loc[0]:[1,2,50],loc[1]:[2,1,49],loc[2]:[50,49,1]}
        my_df = pd.DataFrame(my_data,index=loc)
        dataMD = [
            'distanceMatrix',
            'no file_name',
            'pandas df',
            '1',
            'distance matrix',
            '3',
            '3',
            'utf-8'
        ]
        myNewDB.addDataSet2ExistingFolder(newFolderMD[0],dataMD,my_df)

        # adding new md schema to dataset (user gets error, because at least
        # 1 dataset is stored to database
        myNewDB.add2TemplateMDofDataset(['Creator','time of creation'])

        # adding new md schema to dataset, being sure
        myNewDB.add2TemplateMDofDataset(
            ['Creator','time of creation'],
            sure=True
        )

        # get md info about dataset
        print(
            myNewDB.getMDFromDataSet(
                folder='newfolder',
                dataset='distanceMatrix'
            )
        )

        # editing dataset
        myNewDB.editMDofDataset(
            folder='newfolder',
            dataset='distanceMatrix',
            category='Creator',
            content='Marcel'
        )
        myNewDB.editMDofDataset(
            folder='newfolder',
            dataset='distanceMatrix',
            category='time of creation',
            content='today'
        )


        =======================================================================
    
        '''
        with h5py.File(self.h5file_name, 'a') as db:
            if len(list(db.keys())) > 0:
                if folder in db.keys():
                    if dataset in db[folder].keys():
                        if db[folder].attrs.__contains__(category):
                            if db[
                                folder +
                                '/' +
                                dataset
                            ].attrs.__contains__(category):
                                # edit folder md entry in list
                                where_it_belongs = list(
                                    db[folder].attrs['Dataset']
                                ).index(dataset)
                                db[folder].attrs[category] = \
                                    list(
                                        db[folder].attrs[category]
                                    )[where_it_belongs]
                                # edit dataset md
                                db[
                                    folder +
                                    '/' +
                                    dataset
                                    ].attrs[category] = content
                            else:
                                raise Exception(
                                    'category does not exist in dataset!'
                                )
                        else:
                            raise Exception(
                                'category does not exist in folder!'
                            )
                    else:
                        raise Exception('dataset does not exist!')
                else:
                    raise Exception('folder does not exist!')
            else:
                raise Exception('no folders yet!')

            if isinstance(category, str) and isinstance(content, str):
                if db.attrs.__contains__(category):
                    db.attrs[category] = content
            elif isinstance(category, list) and isinstance(content, list):
                for it, el in enumerate(category):
                    if db.attrs.__contains__(el):
                        db.attrs[el] = content[it]
            else:
                raise Exception(
                    'Both \'category\' and \'content\' must be either str OR '
                    'list. No mixing up please!'
                )

    # TODO test
    # TODO write self test to test new db sync functionality
    # TODO check if ID is unique
    # TODO name shouldnt be editable or sync with db
    # TODO:
    #   Future updates may have the option of having the variable dataset as a
    #   list and content as a list, so that multiple entries may be done at
    #   the same time.
    # TODO:
    #   Is a flag for overwriting existing content necessary?
    def editMDofDB(
            self,
            folder,
            category: str,
            content
    ):
        """
        purpose:
        Adding a new category to the MD schema of the database is done by
        the function add2TemplateMDofDB (see example).
        Using this function only editing of EXISTING md categories is
        supported.
        Adding a new category means an update to md data of database, while
        This function only edits (or adds) content to 1 entry of the database
        at a time.

        TODO:
        Future updates may have the option of having the variable dataset as
        a list and content as a list, so that multiple entries may be done
        at the same time.

        TODO:
        Is a flag for overwriting existing content necessary?


        input:
        :param folder: the folder for which the content is for (str or list)
        :param category: existing categories for changing (str) - 1 category
        at a time
        :param content: corresponding content existing category and
        (str or list of) folders for changing (str or list) - same dimension
        as folder

        :return:

        example:
        from supplychainmodelhelper import mds
        myNewDB = mds.Datacube('test.hdf5','new')

        # add 3 folders
        folder1 = ['test','1','a title','a description']
        folder2 = ['anothertest','2','another title','a new description']
        folder3 = [
            'yetanothertest',
            '3',
            'yet another title',
            'also new description'
        ]
        myNewDB.addFolder2ExistingDB(folder1)
        myNewDB.addFolder2ExistingDB(folder2)
        myNewDB.addFolder2ExistingDB(folder3)

        # add categories without being sure (gets an error, because user
        wasnt sure)
        myNewDB.add2TemplateMDofDB(['source','year','help'])

        # add categories with being sure
        myNewDB.add2TemplateMDofDB(['source','year','help'],sure=True)

        # edit new entries
        folder_names = [folder1[0],folder2[0],folder3[0]]
        myContent = [
            'Astronomische Nachrichten',
            'Zentralblatt Mathematik',
            'FMJ'
        ]
        myNewDB.editMDofDB(
            folder=folder_names,
            category='source',
            content=myContent
        )
        # edit other entries
        myContent = ['2009','2021','2017']
        myNewDB.editMDofDB(
            folder=folder_names,
            category='year',
            content=myContent
        )

        #see changes
        print(myNewDB.getMDfrom_db())




        =======================================================================

        """
        existing_folders = list(self.getMDfrom_db()['Folder'])
        with h5py.File(self.h5file_name, 'a') as db:
            # check if folder is str and content is str
            # as they are to be connected
            if isinstance(folder, str) and isinstance(content, str):
                # if this attribute exists in database, go ahead
                if db.attrs.__contains__(category) \
                        and folder in existing_folders:
                    # get attribute of category and
                    # look up the index where folder is
                    existing_content = db.attrs[category]
                    my_folder_index = list(db.attrs['Folder']).index(folder)
                    # replace existing content with new content
                    existing_content[my_folder_index] = content
                    # store replacement in attributes of database
                    db.attrs[category] = existing_content
                else:
                    raise Exception(
                        'Either the category' +
                        category +
                        ' or the folder' +
                        folder +
                        ' could not be found in database!'
                    )
            elif isinstance(folder, list) and isinstance(content, list):
                if len(folder) != len(content):
                    raise Exception(
                        'both list of folders and list of '
                        'content need to be the same length!'
                    )
                for it, el in enumerate(content):
                    if db.attrs.__contains__(category) and \
                            folder[it] in existing_folders:
                        # get attribute of category and
                        # look up the index where folder is
                        existing_content = list(db.attrs[category])
                        if len(existing_content) == 0:
                            raise Exception(
                                str(category) +
                                ' should result in a list with at least one '
                                'entry. Error in database md schema!'
                            )
                        my_folder_index = list(
                            db.attrs['Folder']
                        ).index(folder[it])
                        # replace existing content with new content
                        existing_content[my_folder_index] = el
                        # store replacement in attributes of database
                        db.attrs[category] = existing_content
                    else:
                        raise Exception(
                            'Either category ' +
                            category +
                            ' or folder ' +
                            str(folder[it]) +
                            ' could not be found in database!'
                        )
            else:
                raise Exception(
                    'Both \'category\' and \'content\' must be either str '
                    'OR list. No mixing up please!'
                )

    # TODO test
    def exportTemplateMDSchemaofDB2CSV(
            self,
            filepath_db_schema
    ):
        '''
        purpose:
        store the current metadata schema of the database as a
        template csv file, for later import
        creates a template DataFrame csv file
        containing minimum mandatory metadata information
        which can be filled in csv file 
        and read in via importMDSchemaofDBfromCSV


        input:
        :param filepath_db_schema: file path to where the template is 
        stored to a csv file

        :return: none

        example: 
        from supplychainmodelhelper import mds
        myNewDB = mds.Datacube('test.hdf5','new')

        myNewDB.exportTemplateMDSchemaofDB2CSV('myBasicMDschemaofDB.csv')

        myNewDB.add2TemplateMDofDB([
            'a new category',
            'another new one',
            'not another one of these'
        )

        myNewDB.exportTemplateMDSchemaofDB2CSV('expandedMDschemaofDB.csv')



        =======================================================================

        '''
        self.filepath_db_schema = filepath_db_schema
        import csv
        with open(self.filepath_db_schema, 'w') as csvfile:
            temp_writer = csv.writer(csvfile, delimiter=';')
            temp_writer.writerow(self.listOfTemplateMDofDB)

    # TODO test
    def exportTemplateMDSchemaofDataset2CSV(
            self,
            filepath_folder_schema
    ):
        '''
        purpose:
        stores the current metadata schema into an csv file.
        information about dataset may be added and later import in via
        importMDSchemaofDatasetFromCSV.


        input:
        :param filepath_folder_schema: file path to where the template is 
        stored to a csv file

        :return: none

        example: 
        from supplychainmodelhelper import mds
        myNewDB = mds.Datacube('test.hdf5','new')

        myNewDB.createTemplate_MD_Schema_Folder_CSV(
            'myBasicMDschemaforAllDatasets.csv'
            )

        # adding more categories
        myNewDB.add2TemplateMDofDataset(['something old','something new'])

        myNewDB.createTemplate_MD_Schema_Folder_CSV(
        'myBasicMDschemaforAllDatasets.csv'
        )


        =======================================================================
        '''
        self.filepath_folder_schema = filepath_folder_schema
        import csv
        with open(self.filepath_folder_schema, 'w') as csvfile:
            temp_writer = csv.writer(csvfile, delimiter=';')
            temp_writer.writerow(self.list_template_md_dataset)

    # TODO test in combination with other stuff
    # TODO check if add2folder doesnt mess with sync the db
    # TODO check if md schema is in agreement with db loaded
    # TODO check if ID is unique
    # TODO check if user imports new md data into existing db -> use case?
    # TODO set flag sure, for user that wants to overwrite current md schema
    def importMDDataofDBfromCSV(
            self,
            csv_file_name
    ):
        '''
        purpose:
        User may export current md schema with function
        exportTemplateMDSchemaofDB2CSV to a csv file, then
        edit csv file and then import the content into the
        Datacube.
        NOTE that database should be freshly initialised and NOT
        contain any other data than an extended md schema up to this point.
        This case has not been tested.

        MD schema of csv and current md schema of db (via add2TemplateMDofDB)
        need to be identical (returns error if doesnt fit).
        I.e. the first row of the csv file should contain the
        fields of the md schema.
        All folders are to be created in the database. The name
        and the ID need to be unique (is enforced!).

        User may want to import self created csv:
        If csv file is not created by exportTemplateMDSchemaofDB2CSV
        the separator used should be a ';'.

        input:
        :param csv_file_name: file_name of the csv file, filled in by user

        :return: 

        example: 
        from supplychainmodelhelper import mds
        myNewDB = mds.Datacube('test.hdf5','new')

        myNewDB.exportTemplateMDSchemaofDB2CSV('file_name.csv')

        #...
        # user sees this in 'file_name.csv'
        'Folder';'ID';'Title';'Description'
        #...
        # user fills in the fields : csv mock up of 'file_name.csv'
        'Folder';'ID';'Title';'Description'
        'test1';'1';'my first Title';'my first description'
        'test2';'2';'my second Title';'my second description'
        'test3';'3';'my third Title';'my third description'
        'test4';'4';'my fourth Title';'my fourth description'

        #...

        myNewDB.importMDSchemaofDBfromCSV('file_name.csv')

        # show me the new folders incl. md schema
        print(myNewDB.getMDfrom_db())




        =======================================================================
    
        '''
        my_table = pd.read_csv(csv_file_name, delimiter=';')

        for row in my_table.index:
            currentlist_entries = []
            for el in self.listOfTemplateMDofDB:
                currentlist_entries.append(my_table[el][row])
            self.addFolder2ExistingDB(currentlist_entries)

    # TODO test
    # TODO write self test to test new db sync functionality
    # TODO check if add2folder doesnt mess with sync the db
    # TODO check if md schema is in agreement with db loaded
    # TODO check if ID is unique
    # TODO if folder doesnt exist in db, create folder!
    def importMDDataofDatasetFromCSV(
            self,
            folder_name,
            csv_file_name
    ):
        '''
        purpose:
        User may export current md schema with function
        exportTemplateMDSchemaofDataset2CSV to a csv file, then
        edit csv file and then import the content into the
        Datacube.

        Checks if current md schema is compatible with imported csv.
        Checks if input parameter folder_name exists as folder in database.
        This function imports all assiociated files named in respective
        md schema and imports them automatically into the database.
        Search for files either in directory where csv file with md data is
        stored or in a subdirectory named 'folder_name'. Returns error if
        neither option is good.

        User may want to import self created csv:
        If csv file is not created by exportTemplateMDSchemaofDB2CSV
        the separator used should be a ';'.
        


        input:
        :param folder_name: name of existing folder in database
        if data files are stored in subdirectory, it should have
        the same name
        :param file_name: file path to where the template is 
        stored to a csv file. file_names given in md schema should
        match the file_names on disk.

        :return: dataframe to check given input

        example: 
        from supplychainmodelhelper import mds
        myNewDB = mds.Datacube('test.hdf5','new')

        myNewDB.addFolder2ExistingDB(['myfolder'])

        myNewDB.exportTemplateMDSchemaofDataset2CSV('file_name.csv')

        print(myNewDB.

        #...
        # user sees this in 'file_name.csv'
        'Dataset';'file_name';'Format';'ID';'Title';'Rows';'Columns';'Encoding'
        # ...
        # user fills in information
        'Dataset';'file_name';'Format';'ID';'Title';'Rows';'Columns';'Encoding'
        'data1';'data1.csv';'csv';'1';'my first title';'4';'3';'utf-8'
        'data2';'data2.csv';'csv';'2';'my new title';'12';'31';'utf-8'
        'data3';'data3.csv';'csv';'3';'my other title';'123';'13';'utf-8'
        # ...

        myNewDB.importMDDataofDatasetFromCSV('myfolder','file_name.csv')
        print(myNewDB.getMDFromFolder('myfolder'))



        =======================================================================

        '''
        my_table = pd.read_csv(csv_file_name, delimiter=';')
        csv_dir = os.path.dirname(os.path.abspath(csv_file_name))

        # check if given folder_name exists
        db_md = self.getMDfrom_db()
        existing_folders = [db_md['Folder'][i] for i in db_md.index]
        if not folder_name in existing_folders:
            raise Exception(
                'Please use function \'addFolder2ExistingDB\' '
                'to add folder with metadata schema to database!'
            )

        expected_files = []
        for myFile in my_table.index:
            expected_files.append(my_table['file_name'][myFile])

        # check if dataset file_names in folder_name are consistent with given md
        # sf -> subfolder (if nothing found in same directory,
        # subfolder folder_name is searched)
        if not os.path.isdir(os.path.join(csv_dir, folder_name)):
            sf = False
            warnings.warn(
                'Cant find subfolder ' +
                str(folder_name) +
                '. Lets check if the files are in the same '
                'folder as the given csv file...'
            )

            # checking as warning indicates
            if not all(
                    [
                        True if i in os.listdir(csv_dir)
                        else False for i in expected_files
                    ]
            ):
                not_found = [
                    i if i not in os.listdir(csv_dir)
                    else 'Found' for i in expected_files
                ]
                raise Filenot_foundError(
                    'Report on missing files: ' +
                    str(not_found)
                )
        else:
            sf = True

        # check if headers of table are concistent with current md schema
        all_found = [True if i in list(my_table.columns)
                     else False for i in self.list_template_md_dataset
                     ]
        if not all_found:
            raise Exception(
                'Did not find all metadata columns needed '
                'for current md schema, please check!'
            )

        # start reading in files:
        for row in my_table.index:
            currentlist_entries = []
            current_file_name = my_table['file_name'][row]
            if sf:
                my_dataset = pd.read_csv(
                    os.path.join(csv_dir, folder_name, current_file_name),
                    ';'
                )
            else:
                my_dataset = pd.read_csv(current_file_name, ';')
            for el in self.list_template_md_dataset:
                currentlist_entries.append(my_table[el][row])
            print('nothing')
            self.addDataSet2ExistingFolder(
                folder_name,
                currentlist_entries,
                my_dataset
            )

    # TODO everything
    def exportSchemaDF2CSV(
            self,
            folder_path,
            df_md
    ):
        """
        purpose:
        save csv file to disk
        create folders named in columns "Folder"
        create corresponding folder metadata files

        input:
        :param xxx:

        :return:

        example:



        =======================================================================

        """
        self.folder_path = folder_path
        self.df_md = df_md

    # TODO subfolder!
    def addFolder2ExistingDB(
            self,
            list_entries
    ):
        '''
        purpose:
        if md schema exists, add row to table
        create folder md template from column folder
        add md schema to db for all folders

        input:
        :param list_entries: list of meta data information. 
        check mandatory fields via listOfTemplateMDofDB

        :return: none

        example:
        from supplychainmodelhelper import mds

        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname,'new')

        print(testDC.listOfTemplateMDofDB)
        >>['Folder','ID','Title','Description']

        # add a folder to your database
        my_list2FillOut = [
            'newfolder',
            '1',
            'This is a test',
            'some kind of description text to describe ' +
            'the reason for the existence of this folder'
            ]
        testDC.addFolder2ExistingDB(list_entries=my_list2FillOut)
   


        =======================================================================
       
        '''
        if len(list_entries) != len(self.listOfTemplateMDofDB):
            raise Exception(
                'given list of entries incompatible '
                'with current metadata schema(' +
                str(len(list_entries)) +
                ' vs. ' +
                str(len(self.listOfTemplateMDofDB)) +
                ')!'
            )

        if list_entries[0] != list_entries[0].replace(" ", ""):
            list_entries[0] = list_entries[0].replace(" ", "")
            warnings.warn(
                'Please be advised, no whitespaces in folder_names! '
                '\nRemoving whitespaces before entering into database'
            )

        with h5py.File(self.h5file_name, 'a') as db:
            # check if folder exists already!
            list_existing_folders = list(db.keys())
            folder_exists = bool(
                set(
                    [list_entries[0]]
                ).intersection(
                    set(list_existing_folders)
                )
            )
            if not folder_exists:
                db.create_group(list_entries[0])
                for index, run in enumerate(self.listOfTemplateMDofDB):
                    # for each folder 1 entry exists
                    # if this is the first entry
                    if not db.attrs.__contains__(run):
                        my_list = []
                    else:  # otherwise just add to the list
                        my_list = list(db.attrs.__getitem__(run))
                    my_list.append(list_entries[index])
                    db.attrs[run] = my_list
            else:
                raise Exception('This folder already exists!')

    # TODO add error of user catches
    # TODO test
    # TODO write self test
    # TODO test what happens if data is in the folder? if everythin is deleted,
    #  add sure flag
    def removeFolder(
            self,
            folder_name: str
    ):
        '''
        purpose:
        removes an existing folder from database.

        input:
        :param folder_name: the name of the folder in the db
        (first entry in md schema)

        :return: none

        example:
        from supplychainmodelhelper import mds

        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname,'new')

        # add a folder to your database
        my_list2FillOut = [
            'newfolder',
            '1',
            'This is a test',
            'some kind of description text to describe the ' +
            'reason for the existence of this folder'
            ]
        testDC.addFolder2ExistingDB(list_entries=my_list2FillOut)

        testDC.getMDfrom_db()

        testDC.removeFolder('newfolder')

        testDC.getMDfrom_db()


        =======================================================================
        '''
        with h5py.File(self.h5file_name, "a") as db:
            db.__delitem__(folder_name)
            fold_ind = list(db.attrs['Folder']).index(folder_name)
            for attr in db.attrs.keys():
                my_list = list(db.attrs[attr])
                del my_list[fold_ind]
                db.attrs[attr] = my_list

        print('folder and corresponding meta data removed')

    # TODO test
    # TODO subfolder! 
    # TODO check axis are realy attached to correct dataframes!
    def addDataSet2ExistingFolder(
            self,
            folder_name,
            list_data_entries,
            dataset_df
    ):
        '''
        purpose: 
        if md schema exists, add row to metadata table in folder
        add dataset to hdf5, reference to name of Dataset
        add metadata scheme of dataset to folder md schema
        NOTE that only 2D dataframes are tested to be stored in this DB!
        NOTE that category 'ID' needs to be unique in this Folder,
        otherwise returns error!

        input:
        :param folder_name: name of an existing folder within hdf5 database
        :param list_data_entries: list of metadata information of current
        md schema
        :param dataset_df: a pandas dataframe with information about axis

        :return: 
        
        example:
        from supplychainmodelhelper import mds
        import pandas as pd

        loc = ['BER','SXF','TXL']

        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname,'new')

        # creating a folder
        my_list2FillOut = [
            'newfolder',
            '1',
            'This is a test',
            'some kind of description text to describe the reason ' +
            'for the existence of this folder'
        ]
        testDC.addFolder2ExistingDB(list_entries=my_list2FillOut)

        # Creating dataset
        my_data={loc[0]:[1,2,50],loc[1]:[2,1,49],loc[2]:[50,49,1]}
        my_df = pd.DataFrame(my_data,index=loc)

        # check current md schema
        print(testDC.list_template_md_dataset)
        >>['Dataset','file_name','Format','ID','Title','Rows','Columns',
        'Encoding']

        # add dataset to an existing folder
        my_list2FillOut =
            'distancematrix',
            'distmatrix.csv',
            'csv',
            '1',
            'Distance matrix',
            str(len(list(my_df.index))),
            str(len(list(my_df.columns))),
            'utf-8'
        ]
        testDC.addDataSet2ExistingFolder(
            folder_name='newfolder',
            list_data_entries=my_list2FillOut,
            dataset_df=my_df
        )



        =======================================================================
        
        '''
        self.list_data_entries = list_data_entries
        self.folder_name = folder_name
        self.dataset_df = dataset_df
        if len(self.list_data_entries) != len(self.list_template_md_dataset):
            raise Exception(
                'given list of entries incompatible '
                'with current metadata schema!'
            )

        if list_data_entries[0] != list_data_entries[0].replace(" ", ""):
            list_data_entries[0] = list_data_entries[0].replace(" ", "")
            warnings.warn(
                'Please be advised, no whitespaces in dataset_names! '
                'Removing whitespaces before entering into database'
            )

        is_there_an_object = [
            True if i == object
            else False for i in list(dataset_df.dtypes)
        ]
        if any(is_there_an_object):
            if is_there_an_object[0]:
                dataset_df.index = dataset_df[list(dataset_df.columns)[0]]
                del dataset_df[list(dataset_df.columns)[0]]
            else:
                raise Exception(
                    'currently only numbers are accepted in this database'
                )

        # create dataset
        # add list of datasets with metadata to database
        # attach metadata of dataset to folder
        # list_template_md_dataset = [
        #   'Dataset'-0,
        #   'file_name'-1,
        #   'Format'-2,
        #   'ID'-3,
        #   'Title'-4,
        #   'Rows'-5,
        #   'Columns'-6,
        #   'Encoding'-7
        #  ]
        with h5py.File(self.h5file_name, 'a') as db:
            # check if folder exists
            list_existing_folders = list(db.keys())
            folder_exists = bool(
                set([folder_name]).intersection(set(list_existing_folders)))

            if folder_exists:
                # adding dataset to hdf5 file
                name_dataset = list_data_entries[0]
                # check if name of data set exists already in this folder
                list_existing_datasets_in_folder = list(db[folder_name].keys())
                dataset_exists = bool(set([name_dataset]).intersection(
                    set(list_existing_datasets_in_folder)))
                # id checker
                if db[folder_name].attrs.__contains__('ID'):
                    list_existing_ids_in_folder = list(
                        db[folder_name].attrs['ID']
                    )
                    id_dataset = list_data_entries[3]
                    id_exists = bool(
                        set([id_dataset]).intersection(
                            set(list_existing_ids_in_folder)
                        )
                    )
                else:
                    id_exists = False

                if id_exists:
                    raise Exception('Please choose another ID!')

                # start entering data after checking
                # if folder exists and dataset is new
                if not dataset_exists:
                    # store data set in db
                    my_d = db[folder_name].create_dataset(
                        name=name_dataset,
                        data=dataset_df
                    )

                    # store md attached to dataset in db
                    for index, run in enumerate(
                            self.list_template_md_dataset):
                        # adding metadata to hdf5 file attached to dataset
                        db[
                            folder_name +
                            '/' +
                            name_dataset
                            ].attrs[run] = self.list_data_entries[index]

                        # adding metadata schema to folder
                        # if first entry in this folder, create list
                        if not db[folder_name].attrs.__contains__(run):
                            my_list = []
                        else:  # otherwise add to existing list
                            my_list = list(
                                db[folder_name].attrs.__getitem__(run))
                        my_list.append(self.list_data_entries[index])
                        db[folder_name].attrs[run] = my_list

                    # store axis information of dataframe attached to dataset
                    try:
                        dt_index = 'i'
                        db[self.folder_name].create_dataset(
                            "ScaleRowOfDataSet" +
                            str(list_data_entries[3]),
                            data=self.dataset_df.index.astype(int).values,
                            dtype=dt_index
                        )
                    except TypeError:
                        dt_index = h5py.string_dtype(encoding='utf-8')
                        db[self.folder_name].create_dataset(
                            "ScaleRowOfDataSet" + str(list_data_entries[3]),
                            data=self.dataset_df.index.values,
                            dtype=dt_index
                        )
                    my_d.dims[0].label = self.list_template_md_dataset[5]
                    my_d.dims[0].attach_scale(
                        db[
                            folder_name +
                            '/ScaleRowOfDataSet' +
                            str(list_data_entries[3])
                            ]
                    )

                    try:
                        dt_col = 'i'
                        db[self.folder_name].create_dataset(
                            "ScaleColOfDataSet" +
                            str(list_data_entries[3]),
                            data=self.dataset_df.columns.astype(int).values,
                            dtype=dt_col
                        )
                    except TypeError:
                        # this only works for py3!
                        dt_col = h5py.string_dtype(encoding='utf-8')
                        db[self.folder_name].create_dataset(
                            "ScaleColOfDataSet" + str(list_data_entries[3]),
                            data=self.dataset_df.columns.values,
                            dtype=dt_col
                        )
                    # encoding list of strings to list of b-strings
                    # (py3 necessity)
                    my_d.dims[1].label = self.list_template_md_dataset[6]
                    my_d.dims[1].attach_scale(
                        db[
                            self.folder_name +
                            "/ScaleColOfDataSet" +
                            str(list_data_entries[3])
                            ]
                    )
                    # list_template_md_dataset = [
                    #   'Dataset'-0,
                    #   'file_name'-1,
                    #   'Format'-2,
                    #   'ID'-3,
                    #   'Title'-4,
                    #   'Rows'-5,
                    #   'Columns'-6,
                    #   'Encoding'-7
                    #  ]
                else:
                    raise Exception(
                        'Name of data set already exists in this '
                        'particular folder, must be unique!'
                    )
            else:
                raise Exception(
                    'This folder does NOT exist. '
                    'Please create a folder by this name!'
                )

    # TODO add error of user catches
    # TODO test
    # TODO write self test now
    def removeDatasetFromFolder(
            self,
            folder_name,
            dataset_name
    ):
        '''
        purpose:
        removes an existing dataset in an existing folder from database.

        input:
        :param folder_name: the name of the folder in the db
        :param dataset_name: the name of the dataset in the folder 'folder_name'

        :return: none

        example:
        from supplychainmodelhelper import mds
        import pandas as pd

        loc = ['BER','SXF','TXL']

        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname,'new')

        # creating a folder
        my_list2FillOut = [
            'newfolder',
            '1',
            'This is a test',
            'some kind of description text to describe the reason' +
            'for the existence of this folder'
        ]
        testDC.addFolder2ExistingDB(list_entries=my_list2FillOut)

        # Creating dataset
        my_data={loc[0]:[1,2,50],loc[1]:[2,1,49],loc[2]:[50,49,1]}
        my_df = pd.DataFrame(my_data,index=loc)

        # check current md schema
        print(testDC.list_template_md_dataset)
        >>['Dataset','file_name','Format','ID','Title','Rows','Columns',
        'Encoding']

        # add dataset to an existing folder
        my_list2FillOut = [
            'distancematrix',
            'distmatrix.csv',
            'csv',
            '1',
            'Distance matrix',
            str(len(list(my_df.index))),
            str(len(list(my_df.columns))),
            'utf-8'
        ]
        testDC.addDataSet2ExistingFolder(
            folder_name='newfolder',
            list_data_entries=my_list2FillOut,
            dataset_df=my_df
        )

        testDC.removeDatasetFromFolder('newfolder','distancematrix')


        =======================================================================
        '''

        with h5py.File(self.h5file_name, "a") as db:
            ds_ind = list(db[folder_name].attrs['Dataset']).index(dataset_name)
            for index, run in enumerate(db[folder_name].attrs.keys()):
                my_list = list(db[folder_name].attrs[run])
                del my_list[ds_ind]
                db[folder_name].attrs[run] = my_list

            del db[folder_name + '/' + dataset_name]
            # list_md_names = list()
        print('dataset removed')

    # TODO test
    def getDataFrameFromFolder(
            self,
            folder_name,
            name_dataset
    ):
        """
        
        purpose: retrieve dataset from hdf5 db as pandas data frame

        input:
        :param folder_name: name of an existing folder within hdf5 database
        :param name_dataset: the name of the dataset given in list of
        md schema (first element of this list)

        :return: the dataframe with axis(if exist)
        
        example:        
        from supplychainmodelhelper import mds

        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname)

         # retrieve existing dataset back from database
         (created by addDataSet2ExistingFolder)
         my_datasetFromhdf5DB = testDC.getDataFrameFromFolder(
            folder_name='newfolder',
            name_dataset='distancematrix'
        )

        =======================================================================
        """
        self.folder_name = folder_name
        self.name_dataset = name_dataset

        with h5py.File(self.h5file_name, 'r') as db:
            list_existing_folders = list(db.keys())
            folder_exists = bool(set([self.folder_name]).intersection(
                set(list_existing_folders)))
            if not folder_exists:
                raise Exception('Folder name doesnt exist. Please check!')
            try:
                from_db = db[folder_name + '/' + name_dataset]
            except KeyError:
                raise Exception('data set with this name is not stored here!')
            my_df = pd.DataFrame(
                data=from_db,
                index=decBSList(list(from_db.dims[0][0])),
                columns=decBSList(list(from_db.dims[1][0]))
            )
        return my_df

    # TODO test
    def getMDFromDataSet(
            self,
            folder_name,
            name_dataset
    ):
        """

        purpose: retrieve metadata from dataset in hdf5 db

        input:
        :param folder_name: name of an existing folder within hdf5 database
        :param name_dataset: the name of the dataset given in list of md schema
        (first element of this list)

        :return: a dataframe(2 column table) of information about an existing
        dataset with headers: MD Names, Content

        example:
        from supplychainmodelhelper import mds

        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname)

         # retrieve existing dataset back from database
         (created by addDataSet2ExistingFolder)
        print(testDC.getMDFromDataSet(
            folder_name='newfolder',
            name_dataset='distancematrix')
            )



        =======================================================================

        """
        self.folder_name = folder_name
        self.name_dataset = name_dataset
        list_md_names = []
        list_md_contents = []

        with h5py.File(self.h5file_name, 'r') as db:
            try:
                for x in db[folder_name +
                            '/' +
                            name_dataset
                ].attrs.__iter__():
                    list_md_names.append(x)
                    list_md_contents.append(
                        db[folder_name +
                           '/' +
                           name_dataset
                           ].attrs.__getitem__(x))
            except KeyError:
                raise Exception('data set with this name is not stored here!')
        return pd.DataFrame(
            data={'MD Names': list_md_names, 'Content': list_md_contents})

    # TODO get like DB otherwise IT WONT WORK!
    def getMDFromFolder(
            self,
            folder_name
    ):
        """
        purpose:
        retrieve metadata from folder, information about the folder in question
        and which information they contain

        input:
        :param folder_name: name of an existing folder within hdf5 database

        :return:  a dataframe(2 column table) of information about an existing
        dataset with headers: MD Names, Content wit list of all datasets

        example:
        from supplychainmodelhelper import mds

        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname)

         # retrieve existing folder back from database
         (created by addDataSet2ExistingFolder)
        print(testDC.getMDFromFolder(folder_name='newfolder'))



        =======================================================================

        """
        self.folder_name = folder_name
        list_md_names = []
        list_md_contents = []

        with h5py.File(self.h5file_name, 'r') as db:
            try:
                for x in db[folder_name].attrs.__iter__():
                    list_md_names.append(x)
                    list_md_contents.append(
                        db[folder_name].attrs.__getitem__(x)
                    )
            except KeyError:
                raise Exception(
                    'folder name with this name is not stored here!'
                )
        return pd.DataFrame(
            data={'MD Names': list_md_names, 'Content': list_md_contents}
        )

    # TODO test with different content and customized metadata schemas
    def getMDfrom_db(self):
        """ 
        purpose:
        retrieve metadata from db, information about folders in this database
        and which information they contain returned as pandas dataframe

        input: no parameters needed, as each instance deals with just one
        database

        :return: a dataframe (multi-column table) of information about an
        existing dataset with headers based on current md schema
        (basic schema headers: Description, Folder, ID, Title)
        
        example:
        from supplychainmodelhelper import mds

        myfname = './testDCnew.hdf5'
        testDC = mds.Datacube(myfname)

         # retrieve md info about database
         (created by addDataSet2ExistingFolder)
        print(testDC.getMDfrom_db())

        ======================================================================
        """
        try:
            with h5py.File(self.h5file_name, 'r') as db:
                my_keys = [key for key in db.attrs.keys()]
                my_vals = [val for val in db.attrs.values()]
        except KeyError:
            raise Exception('data set with this name is not stored here!')

        my_df = pd.DataFrame(columns=my_keys)
        for i, x in enumerate(my_vals):
            my_df[my_keys[i]] = list(my_vals[i])
        return my_df



from setuptools import setup, find_packages


with open("README.md", "r") as readme_file:
    readme = readme_file.read()

with open('requirements_dev.txt') as f:
    requirements = f.read().splitlines()

'''
- Increment the MAJOR version when you make 
incompatible API changes.
- Increment the MINOR version when you add 
functionality in a backwards-compatible manner.
- Increment the PATCH version when you make 
backwards-compatible bug fixes.

'''
setup(
    name="supplychainmodelhelper",
    version="0.2.1",
    author="Marcel Fuhrmann",
    author_email="dr.marcel.fuhrmann@gmail.com",
    description="A package to help with your supply chain model",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/DjMaFu/supplychainmodulator",
    packages=find_packages(),
    install_requires=requirements,
    setup_requires=["pytest-runner"],
    tests_require=["pytest==6.2.2"],
    test_suite='tests',
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)